import itertools
import json

import numpy as np
from bsb import Branch, MissingMorphologyError, Morphology
from bsb import MorphologyRepository as IMorphologyRepository
from bsb import MorphologyRepositoryError, StoredMorphology
from bsb._encoding import EncodedLabels

from .resource import HANDLED, Resource, handles_handles

_root = "/morphologies"


class MetaEncoder(json.JSONEncoder):
    """
    Encodes morphology metadata to JSON
    """

    def default(self, o):
        if isinstance(o, np.ndarray):
            arr = o.tolist()
            arr.append("__ndarray__")
            return arr
        else:
            super().default(o)


def meta_object_hook(obj):
    for k, v in obj.items():
        if isinstance(v, list) and v[-1] == "__ndarray__":
            v.pop()
            obj[k] = np.array(v)
    return obj


class MorphologyRepository(Resource, IMorphologyRepository):
    def __init__(self, engine):
        super().__init__(engine, _root)

    def select(self, *selectors):
        if not selectors:
            return []
        all_loaders = self.all()
        selected = []
        for selector in selectors:
            selector.validate(all_loaders)
            selected.extend(filter(selector.pick, all_loaders))
        return selected

    @handles_handles("r")
    def preload(self, name, meta=None, handle=HANDLED):
        return StoredMorphology(
            name,
            self._make_loader(name, meta),
            meta if meta is not None else self.get_meta(name, handle=handle),
        )

    def _make_loader(self, name, meta):
        def loader():
            return self.load(name, preloaded_meta=meta)

        return loader

    @handles_handles("r")
    def get_meta(self, name, handle=HANDLED):
        all_meta = self.get_all_meta(handle=handle)
        try:
            meta = all_meta[name]
        except KeyError:
            raise MissingMorphologyError(
                f"`{self._engine.root}` contains no morphology named `{name}`."
            ) from None
        return meta

    @handles_handles("r")
    def get_all_meta(self, handle=HANDLED):
        if "morphology_meta" not in handle:
            return {}
        return json.loads(handle["morphology_meta"][()], object_hook=meta_object_hook)

    @handles_handles("a")
    def update_all_meta(self, meta, handle=HANDLED):
        all_meta = self.get_all_meta(handle=handle)
        all_meta.update(meta)
        self.set_all_meta(all_meta, handle=handle)

    @handles_handles("a")
    def set_all_meta(self, all_meta, handle=HANDLED):
        if "morphology_meta" in handle:
            del handle["morphology_meta"]
        handle.create_dataset(
            "morphology_meta", data=json.dumps(all_meta, cls=MetaEncoder)
        )

    @handles_handles("r")
    def all(self, handle=HANDLED):
        all_meta = self.get_all_meta(handle=handle)
        return [
            self.preload(name, meta=meta, handle=handle)
            for name, meta in all_meta.items()
        ]

    @handles_handles("r")
    def has(self, name, handle=HANDLED):
        return f"{self._path}/{name}" in handle

    @handles_handles("r")
    def load(self, name, preloaded_meta=None, handle=HANDLED):
        try:
            root = handle[f"{self._path}/{name}/"]
        except Exception:
            raise MissingMorphologyError(
                f"`{self._engine.root}` contains no morphology named `{name}`."
            ) from None
        data = root["data"][()]
        points = data[:, :3].copy()
        radii = data[:, 3].copy()
        # Turns the forced JSON str keys back into ints
        labelsets = {
            int(k): v for k, v in json.loads(root["data"].attrs["labels"]).items()
        }
        labels = EncodedLabels(
            len(points), buffer=data[:, 4].astype(int), labels=labelsets
        )
        prop_names = root["data"].attrs["properties"]
        props = dict(zip(prop_names, np.rollaxis(data[:, 5:], 1)))
        parents = {-1: None}
        branch_id = itertools.count()
        roots = []
        ptr = 0
        for nptr, p in root["graph"][()]:
            radii[ptr:nptr]
            labels[ptr:nptr]
            {k: v[ptr:nptr] for k, v in props.items()}
            branch = Branch(
                points[ptr:nptr],
                radii[ptr:nptr],
                labels[ptr:nptr],
                {k: v[ptr:nptr] for k, v in props.items()},
            )
            parent = parents.get(p, None)
            parents[next(branch_id)] = branch
            if parent:
                parent.attach_child(branch)
            else:
                roots.append(branch)
            ptr = nptr
        if preloaded_meta is None:
            meta = self.get_meta(name, handle=handle)
        else:
            meta = preloaded_meta
        morpho = Morphology(roots, meta, shared_buffers=(points, radii, labels, props))
        assert morpho._check_shared(), "Morpho read with unshareable buffers"
        return morpho

    @handles_handles("a")
    def save(self, name, morphology, overwrite=False, update_meta=True, handle=HANDLED):
        me = handle[self._path]
        if self.has(name):
            if overwrite:
                self.remove(name)
            else:
                root = self._engine.root
                raise MorphologyRepositoryError(
                    f"A morphology called '{name}' already exists in `{root}`."
                )
        root = me.create_group(name)
        # Optimizing a morphology goes through the same steps as what is required
        # to save it to disk; plus, now the user's object is optimized :)
        morphology.optimize()
        branches = morphology.branches
        n_prop = len(morphology._shared._prop)
        data = np.empty((len(morphology), 5 + n_prop))
        data[:, :3] = morphology._shared._points
        data[:, 3] = morphology._shared._radii
        data[:, 4] = morphology._shared._labels
        for i, prop in enumerate(morphology._shared._prop.values()):
            data[:, 5 + i] = prop
        dds = root.create_dataset("data", data=data)
        dds.attrs["labels"] = json.dumps(
            {k: list(v) for k, v in morphology._shared._labels.labels.items()}
        )
        dds.attrs["properties"] = [*morphology._shared._prop.keys()]
        graph = np.empty((len(branches), 2))
        parents = {None: -1}
        ptr = 0
        for i, branch in enumerate(morphology.branches):
            ptr += len(branch)
            graph[i, 0] = ptr
            graph[i, 1] = parents[branch.parent]
            parents[branch] = i
        root.create_dataset("graph", data=graph, dtype=int)
        morphology.meta["name"] = name
        if len(morphology._shared._points):
            morphology.meta["ldc"] = np.min(morphology._shared._points, axis=0)
            morphology.meta["mdc"] = np.max(morphology._shared._points, axis=0)
        else:
            morphology.meta["ldc"] = morphology.meta["mdc"] = np.nan
        if update_meta:
            all_meta = self.get_all_meta(handle=handle)
            all_meta[name] = morphology.meta
            self.set_all_meta(all_meta)
        return StoredMorphology(name, lambda: morphology, morphology.meta)

    @handles_handles("a")
    def remove(self, name, handle=HANDLED):
        try:
            del handle[f"{self._path}/{name}"]
        except KeyError:
            raise MorphologyRepositoryError(f"'{name}' doesn't exist.") from None



"""
HDF5 storage engine for the BSB framework.
"""

import json
import os
import shutil
from datetime import datetime

import h5py
import shortuuid
from bsb import Engine, MPILock
from bsb import StorageNode as IStorageNode
from bsb import __version__ as bsb_version
from bsb import config, report

from .connectivity_set import ConnectivitySet
from .file_store import FileStore
from .morphology_repository import MorphologyRepository
from .placement_set import PlacementSet

__version__ = "5.0.1"
__all__ = [
    "PlacementSet",
    "ConnectivitySet",
    "FileStore",
    "MorphologyRepository",
    "HDF5Engine",
    "StorageNode",
]


def on_main(prep=None, ret=None):
    def decorator(f):
        def wrapper(self, *args, **kwargs):
            r = None
            self.comm.barrier()
            if self.comm.get_rank() == 0:
                r = f(self, *args, **kwargs)
            elif prep:
                prep(self, *args, **kwargs)
            self.comm.barrier()
            if not ret:
                return self.comm.bcast(r, root=0)
            else:
                return ret(self, *args, **kwargs)

        return wrapper

    return decorator


def on_main_until(until, prep=None, ret=None):
    def decorator(f):
        def wrapper(self, *args, **kwargs):
            global _procpass
            r = None
            self.comm.barrier()
            if self.comm.get_rank() == 0:
                r = f(self, *args, **kwargs)
            elif prep:
                prep(self, *args, **kwargs)
            self.comm.barrier()
            while not until(self, *args, **kwargs):
                pass
            if not ret:
                return self.comm.bcast(r, root=0)
            else:
                return ret(self, *args, **kwargs)

        return wrapper

    return decorator


def _set_root(self, root):
    self._root = root


def _set_active_cfg(self, config):
    config._meta["active_config"] = True


class NoopLock:
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class HDF5Engine(Engine):
    def __init__(self, root, comm):
        super().__init__(root, comm)
        self._lock = MPILock.sync(comm._comm)
        self._readonly = False

    @on_main()
    @property
    def versions(self):
        with self._handle("r") as handle:
            return {
                "bsb": handle.attrs["bsb_version"],
                "engine": "bsb-hdf5",
                "version": handle.attrs["bsb_hdf5_version"],
            }

    @property
    def root_slug(self):
        return os.path.relpath(self._root)

    @staticmethod
    def recognizes(root, comm):
        lock = MPILock.sync(comm)
        with lock.read():
            try:
                h5py.File(root, "r").close()
                return True
            except Exception as e:
                report(f"Cannot open hdf5 file {root}: {e}.", level=1)
                return False

    def _read(self):
        # h5py does not deal with concurrent read access
        return self._lock.read()

    def _write(self):
        if self._readonly:
            raise IOError("Can't perform write operations in readonly mode.")
        else:
            return self._lock.write()

    def _master_write(self):
        if self._readonly:
            raise IOError("Can't perform write operations in readonly mode.")
        else:
            return self._lock.single_write()

    def _handle(self, mode):
        if self._readonly and mode != "r":
            raise IOError("Can't perform write operations in readonly mode.")
        else:
            return h5py.File(self._root, mode)

    def exists(self):
        return os.path.exists(self._root)

    @on_main_until(lambda self: self.exists())
    def create(self):
        with self._handle("w") as handle:
            handle.attrs["bsb_hdf5_version"] = __version__
            handle.attrs["bsb_version"] = bsb_version
            handle.create_group("placement")
            handle.create_group("connectivity")
            handle.create_group("files")
            handle.create_group("morphologies")

    @on_main_until(lambda self, r: self.exists(), _set_root)
    def move(self, new_root):
        shutil.move(self._root, new_root)
        self._root = new_root

    @on_main_until(lambda self, r: self.__class__(self.root, self.comm).exists())
    def copy(self, new_root):
        shutil.copy(self._root, new_root)

    @on_main_until(lambda self: not self.exists())
    def remove(self):
        os.remove(self._root)

    @on_main_until(
        lambda self, ct: PlacementSet.exists(self, ct),
        prep=None,
        ret=lambda self, ct: PlacementSet(self, ct),
    )
    def require_placement_set(self, ct):
        return PlacementSet.require(self, ct)

    @on_main()
    def clear_placement(self):
        with self._handle("a") as handle:
            handle.require_group("placement")
            del handle["placement"]
            del handle.attrs["chunk_size"]
            handle.require_group("placement")
            self._write_chunk_stats(handle, {})

    @on_main()
    def clear_connectivity(self):
        with self._handle("a") as handle:
            handle.require_group("connectivity")
            del handle["connectivity"]
            handle.require_group("connectivity")
            stats = self._read_chunk_stats(handle)
            stats = {
                k: {"placed": v["placed"], "connections": {"inc": 0, "out": 0}}
                for k, v in stats.items()
            }
            self._write_chunk_stats(handle, stats)

    @on_main()
    def get_chunk_stats(self):
        with self._handle("r") as handle:
            return self._read_chunk_stats(handle)

    def _read_chunk_stats(self, handle: h5py.File) -> object:
        return json.loads(handle.attrs.get("chunks", "{}"))

    def _write_chunk_stats(self, handle, stats):
        handle.attrs["chunks"] = json.dumps(stats)


def _get_default_root():
    return os.path.abspath(
        os.path.join(
            ".",
            "scaffold_network_"
            + datetime.now().strftime("%Y_%m_%d")
            + "_"
            + shortuuid.uuid()
            + ".hdf5",
        )
    )


@config.node
class StorageNode(IStorageNode):
    root = config.attr(type=str, default=_get_default_root, call_default=True)
    """
    Path to the HDF5 network storage file.
    """



import json

import errr
import numpy as np
from bsb import CellType, Chunk
from bsb import ConnectivitySet as IConnectivitySet
from bsb import DatasetNotFoundError, chunklist

from .resource import HANDLED, Resource, handles_handles

_root = "/connectivity/"


class LocationOutOfBoundsError(Exception):
    pass


class ConnectivitySet(Resource, IConnectivitySet):
    """
    Fetches placement data from storage.

    .. note::

        Use :meth:`Scaffold.get_connectivity_set <bsb.core.Scaffold.get_connectivity_set>`
        to correctly obtain a :class:`~bsb.storage.interfaces.ConnectivitySet`.
    """

    pre_type: CellType
    post_type: CellType

    @handles_handles("r", handler=lambda args: args[1])
    def __init__(self, engine, tag, handle=HANDLED):
        self.tag = tag
        self.pre_type = None
        self.post_type = None
        super().__init__(engine, _root + tag)
        if not self.exists(engine, tag, handle=handle):
            raise DatasetNotFoundError(
                f"ConnectivitySet '{tag}' does not exist. Choose from: "
                + errr.quotejoin(self.get_tags(self._engine))
            )
        self.pre_type_name = handle[self._path].attrs["pre"]
        self.post_type_name = handle[self._path].attrs["post"]

    def __len__(self):
        return sum(len(data[0]) for _, _, _, data in self.flat_iter_connections("inc"))

    @classmethod
    @handles_handles("r", handler=lambda args: args[1])
    def get_tags(cls, engine, handle=HANDLED):
        """
        Returns all the connectivity tags in the network.
        """
        return list(handle[_root].keys())

    @classmethod
    @handles_handles("a", handler=lambda args: args[1])
    def create(cls, engine, pre_type, post_type, tag=None, handle=HANDLED):
        """
        Create the structure for this connectivity set in the HDF5 file. Connectivity sets
        are stored under ``/connectivity/<tag>``.
        """
        if tag is None:
            tag = f"{pre_type.name}_to_{post_type.name}"
        path = _root + tag
        g = handle.create_group(path)
        g.attrs["pre"] = pre_type.name
        g.attrs["post"] = post_type.name
        g.require_group(f"{path}/inc")
        g.require_group(f"{path}/out")
        cs = cls(engine, tag, handle=handle)
        cs.pre_type = pre_type
        cs.post_type = post_type
        return cs

    @staticmethod
    @handles_handles("r", handler=lambda args: args[0])
    def exists(engine, tag, handle=HANDLED):
        """
        Checks whether a :class:`~.connectivity_set.ConnectivitySet` with the given tag
        exists.

        :param engine: Engine to use for the lookup.
        :type engine: :class:`.HDF5Engine`
        :param tag: Tag of the set to look for.
        :type tag: str
        :param handle: An open handle to use instead of opening one.
        :type handle: :class:`h5py.File`
        :returns: Whether the tag exists.
        :rtype: bool
        """
        return _root + tag in handle

    @classmethod
    @handles_handles("a", handler=lambda args: args[1])
    def require(cls, engine, pre_type, post_type, tag=None, handle=HANDLED):
        """
        Get or create a :class:`~.connectivity_set.ConnectivitySet`.

        :param engine: Engine to fetch/write the data.
        :type engine: :class:`.HDF5Engine`
        :param pre_type: Presynaptic cell type.
        :type pre_type: :class:`~bsb.cell_types.CellType`
        :param post_type: Postsynaptic cell type.
        :type post_type: :class:`~bsb.cell_types.CellType`
        :param tag: Tag to store the set under. Defaults to
          ``{pre_type.name}_to_{post_type.name}``.
        :type tag: str
        :returns: Existing or new connectivity set.
        :rtype: :class:`~.connectivity_set.ConnectivitySet`
        """
        if tag is None:
            tag = f"{pre_type.name}_to_{post_type.name}"
        path = _root + tag
        g = handle.require_group(path)
        if g.attrs.setdefault("pre", pre_type.name) != pre_type.name:
            raise ValueError(
                "Given and stored type mismatch:"
                + f" {pre_type.name} vs {g.attrs['pre']}"
            )
        if g.attrs.setdefault("post", post_type.name) != post_type.name:
            raise ValueError(
                "Given and stored type mismatch:"
                + f" {post_type.name} vs {g.attrs['post']}"
            )
        g.require_group(path + "/inc")
        g.require_group(path + "/out")
        cs = cls(engine, tag, handle=handle)
        cs.pre_type_name = pre_type.name
        cs.post_type_name = post_type.name
        return cs

    @handles_handles("a")
    def clear(self, handle=HANDLED):
        path = _root + self.tag
        g = handle.require_group(path)
        del g["inc"]
        del g["out"]
        g.create_group("inc")
        g.create_group("out")
        g.attrs["len"] = 0
        g.attrs["chunks"] = "{}"

    @handles_handles("a")
    def connect(self, pre_set, post_set, src_locs, dest_locs, handle=HANDLED):
        src_locs = _point_to_2d(src_locs)
        dest_locs = _point_to_2d(dest_locs)
        if not len(src_locs):
            return
        if len(src_locs) != len(dest_locs):
            raise ValueError("Location matrices must be of same length.")
        if pre_set._requires_morpho_mapping():
            src_locs = pre_set._morpho_backmap(src_locs)
        if post_set._requires_morpho_mapping():
            dest_locs = post_set._morpho_backmap(dest_locs)
        for data in self._demux(pre_set, post_set, src_locs, dest_locs):
            if not len(data[-1]):
                # Don't write empty data
                continue
            self.chunk_connect(*data, handle=handle)

    def _demux(self, pre, post, src_locs, dst_locs):
        src_chunks = pre.get_loaded_chunks()
        lns = []
        for src in iter(src_chunks):
            with pre.chunk_context([src]):
                lns.append(len(pre))
        dmax = 0
        # Iterate over each source chunk
        for dst in post.get_loaded_chunks():
            # Count the number of cells
            with post.chunk_context([dst]):
                ln = len(post)
            dst_idx = dst_locs[:, 0] < ln
            dst_block = dst_locs[dst_idx]
            src_block = src_locs[dst_idx]
            if len(src_chunks) == 1:
                block_idx = np.lexsort((src_block[:, 0], dst_block[:, 0]))
                yield src, dst, src_block[block_idx], dst_block[block_idx]
            else:
                for src, sln in zip(iter(src_chunks), lns):
                    block_idx = (src_block[:, 0] >= 0) & (src_block[:, 0] < sln)
                    yield src, dst, src_block[block_idx], dst_block[block_idx]
                    src_block[:, 0] -= sln
            dst_locs = dst_locs[~dst_idx]
            src_locs = src_locs[~dst_idx]
            # We sifted `ln` cells out of the dataset, so reduce the ids.
            dst_locs[:, 0] -= ln
            dmax += ln
        if len(dst_locs) > 0:
            smax = sum(lns) - 1
            dmax -= 1
            src_msg = ""
            if (m := np.max(src_locs)) > smax:
                src_msg = f"Source maximum is {smax}, but up to {m} given."
            dst_msg = ""
            if (m := np.max(src_locs)) > smax:
                if src_msg:
                    dst_msg = "\n"
                dst_msg += f"Destination maximum is {dmax}, but up to {m} given."
            raise LocationOutOfBoundsError(
                f"Received {len(dst_locs)} out of bounds locations:"
                f"\n- Source locations:\n{src_locs}"
                f"\n- Destinations:\n{dst_locs}\n" + src_msg + dst_msg
            )

    def _store_pointers(self, group, chunk, n, total):
        chunks = [Chunk(t, (0, 0, 0)) for t in group.attrs.get("chunk_list", [])]
        if chunk in chunks:
            # Source chunk already existed, just increment the subseq. pointers
            inc_from = chunks.index(chunk) + 1
        else:
            # We are the last chunk, we start adding rows at the end.
            group.attrs[str(chunk.id)] = total
            # Move up the increment pointer to place ourselves after the
            # last element
            chunks.append(chunk)
            inc_from = len(chunks)
            group.attrs["chunk_list"] = chunks
        # Increment the pointers of the chunks that follow us, by `n` rows.
        for c in chunks[inc_from:]:
            group.attrs[str(c.id)] += n

    def _get_sorted_pointers(self, group):
        chunks = [Chunk(t, (0, 0, 0)) for t in group.attrs["chunk_list"]]
        ptrs = np.array([group.attrs[str(c.id)] for c in chunks])
        sorted = np.argsort(ptrs)
        chunks = [chunks[cid] for cid in sorted]
        return chunks, ptrs[sorted]

    def _get_insert_pointers(self, group, chunk):
        chunks = [Chunk(t, (0, 0, 0)) for t in group.attrs["chunk_list"]]
        iptr = group.attrs[str(chunk.id)]
        idx = chunks.index(chunk)
        if idx + 1 == len(chunks):
            # Last chunk, no end pointer
            eptr = None
        else:
            # Get the pointer of the next chunk
            eptr = group.attrs[str(chunks[idx + 1].id)]
        return iptr, eptr

    @handles_handles("a")
    def chunk_connect(self, src_chunk, dst_chunk, src_locs, dst_locs, handle=HANDLED):
        if len(src_locs) != len(dst_locs):
            raise ValueError("Location matrices must be of same length.")
        self._insert("inc", dst_chunk, src_chunk, dst_locs, src_locs, handle)
        self._insert("out", src_chunk, dst_chunk, src_locs, dst_locs, handle)
        self._track_add(handle, src_chunk, dst_chunk, len(src_locs))

    def _insert(self, tag, local_, global_, lloc, gloc, handle):
        grp = handle.require_group(f"{self._path}/{tag}/{local_.id}")
        unpack_me = [None, None]
        # require_dataset doesn't work for resizable datasets, see
        # https://github.com/h5py/h5py/issues/2018
        # So we create a little thingy for requiring src & dest
        for i, tag in enumerate(("local_locs", "global_locs")):
            if tag in grp:
                unpack_me[i] = grp[tag]
            else:
                unpack_me[i] = grp.create_dataset(
                    tag, shape=(0, 3), dtype=int, chunks=(1024, 3), maxshape=(None, 3)
                )
        lcl_ds, gbl_ds = unpack_me
        # Move the pointers that keep track of the chunks
        new_rows = len(lloc)
        total = len(lcl_ds)
        self._store_pointers(grp, global_, new_rows, total)
        iptr, eptr = self._get_insert_pointers(grp, global_)
        if eptr is None:
            eptr = total + new_rows
        # Resize and insert data.
        lcl_end = lcl_ds[(eptr - new_rows) :]
        gbl_end = gbl_ds[(eptr - new_rows) :]
        lcl_ds.resize(len(lcl_ds) + new_rows, axis=0)
        gbl_ds.resize(len(gbl_ds) + new_rows, axis=0)
        lcl_ds[iptr:eptr] = np.concatenate((lcl_ds[iptr : (eptr - new_rows)], lloc))
        lcl_ds[eptr:] = lcl_end
        gbl_ds[iptr:eptr] = np.concatenate((gbl_ds[iptr : (eptr - new_rows)], gloc))
        gbl_ds[eptr:] = gbl_end

    def _track_add(self, handle, src_chunk, dst_chunk, count):
        # Track addition in global chunk stats
        global_stats = self._engine._read_chunk_stats(handle)
        for tag, chunk in (("inc", dst_chunk), ("out", src_chunk)):
            id = str(chunk.id)
            chunk_stats = global_stats.setdefault(
                id, {"placed": 0, "connections": {"inc": 0, "out": 0}}
            )
            chunk_stats["connections"][tag] += count
            # Track addition in connectivity set
            group = handle[self._path]
            if tag == "out":
                group.attrs["len"] = group.attrs.get("len", 0) + count
            conn_stats = json.loads(group.attrs.get("chunks", "{}"))
            conn_stats.setdefault(id, {"inc": 0, "out": 0})[tag] += count
            group.attrs["chunks"] = json.dumps(conn_stats)
        self._engine._write_chunk_stats(handle, global_stats)

    @handles_handles("r")
    def get_chunk_stats(self, handle=HANDLED):
        return json.loads(handle[self._path].attrs.get("chunks", "{}"))

    @handles_handles("r")
    def get_local_chunks(self, direction, handle=HANDLED):
        return chunklist(
            Chunk.from_id(int(k), None) for k in handle[self._path][direction].keys()
        )

    @handles_handles("r")
    def get_global_chunks(self, direction, local_, handle=HANDLED):
        try:
            chunk_group = handle[self._path][f"{direction}/{local_.id}"]
        except KeyError:
            # The local chunk does not exist, create empty chunklist.
            return chunklist(())
        else:
            # The local chunk exists, return the list of chunks it has data of.
            return chunklist(Chunk(k, None) for k in chunk_group.attrs["chunk_list"])

    def nested_iter_connections(self, direction=None, local_=None, global_=None):
        """
        Iterates over the connectivity data, leaving room for the end-user to set up
        nested for loops:

        .. code-block:: python

          for dir, local_itr in self.nested_iter_connections():
              for lchunk, global_itr in local_itr:
                  print("I can do something at the start of a new local chunk")
                  for gchunk, data in global_itr:
                      print(f"Nested {dir} block between {lchunk} and {gchunk}")
                  print("Or right before we move to the next local chunk")

        If a keyword argument is given, that axis is not iterated over, and the amount of
        nested loops is reduced.

        :param direction: When omitted, iterates ``inc`` and ``out``, otherwise when
          given, pins it to the given value
        :type direction: str
        :param local_: When omitted, iterates over all local chunks in the set. When
          given, it restricts the iteration to the given value(s).
        :type local_: Union[~bsb.storage._chunks.Chunk, list[~bsb.storage._chunks.Chunk]]
        :param global_: When omitted, iterates over all global chunks in the set. When
          given, it restricts the iteration to the given value(s).
        :type global_: Union[~bsb.storage._chunks.Chunk, list[~bsb.storage._chunks.Chunk]]
        :returns: An iterator that produces the next unrestricted iteration values, or
          the connection dataset that matches the iteration combination.
        """
        return CSIterator(self, direction, local_, global_)

    def flat_iter_connections(self, direction=None, local_=None, global_=None):
        """
        Iterates over the connectivity data.

        .. code-block:: python

          for dir, lchunk, gchunk, data in self.flat_iter_connections():
              print(f"Flat {dir} block between {lchunk} and {gchunk}")

        If a keyword argument is given, that axis is not iterated over, and the value is
        fixed in each iteration.

        :param direction: When omitted, iterates ``inc`` and ``out``. When given, it
          restricts the iteration to the given value.
        :type direction: str
        :param local_: When omitted, iterates over all local chunks in the set. When
          given, it restricts the iteration to the given value(s).
        :type local_: Union[~bsb.storage._chunks.Chunk, list[~bsb.storage._chunks.Chunk]]
        :param global_: When omitted, iterates over all global chunks in the set. When
          given, it restricts the iteration to the given value(s).
        :type global_: Union[~bsb.storage._chunks.Chunk, list[~bsb.storage._chunks.Chunk]]
        :returns: Yields the direction, local chunk, global chunk, and data. The data is a
          tuple of the local and global connection locations.
        :rtype: Tuple[str, ~bsb.storage._chunks.Chunk, ~bsb.storage._chunks.Chunk,
          Tuple[numpy.ndarray, numpy.ndarray]]
        """
        itr = CSIterator(self, direction, local_, global_)
        for direction in get_dir_iter(direction):
            for lchunk in itr.get_local_iter(direction, local_):
                for gchunk in itr.get_global_iter(direction, lchunk, global_):
                    conns = self.load_block_connections(direction, lchunk, gchunk)
                    yield direction, lchunk, gchunk, conns

    @handles_handles("r")
    def load_block_connections(self, direction, local_, global_, handle=HANDLED):
        """
        Load the connection block with given direction between the given local and global
        chunk.

        :param direction: Either ``inc`` to load the connections from the incoming
          perspective or ``out`` for the outgoing perspective.
        :type direction: str
        :param local_: Local chunk
        :type local_: ~bsb.storage._chunks.Chunk
        :param global_: Global chunk
        :type global_: ~bsb.storage._chunks.Chunk
        :param handle: This parameter is injected and doesn't have to be passed.
        :returns: The local and global connections locations
        :rtype: Tuple[numpy.ndarray, numpy.ndarray]
        """
        try:
            local_grp = handle[self._path][f"{direction}/{local_.id}"]
            start, end = self._get_insert_pointers(local_grp, global_)
        except KeyError:
            # If a local or global chunk isn't found, return empty data.
            return (np.empty((0, 3), dtype=int), np.empty((0, 3), dtype=int))

        idx = slice(start, end)
        return (local_grp["local_locs"][idx], local_grp["global_locs"][idx])

    @handles_handles("r")
    def load_local_connections(self, direction, local_, handle=HANDLED):
        """
        Load all the connections of the given local chunk.

        :param direction: Either ``inc`` to load the connections from the incoming
          perspective or ``out`` for the outgoing perspective.
        :type direction: str
        :param local_: Local chunk
        :type local_: ~bsb.storage._chunks.Chunk
        :param handle: This parameter is injected and doesn't have to be passed.
        :returns: The local connection locations, a vector of the global connection chunks
          (1 chunk id per connection) and the global connections locations. To identify a
          cell in the global connections, use the corresponding chunk id from the second
          return value.
        :rtype: Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]
        """
        local_grp = handle[self._path][f"{direction}/{local_.id}"]
        global_locs = local_grp["global_locs"][()]
        chunks, ptrs = self._get_sorted_pointers(local_grp)
        col = np.repeat([c.id for c in chunks], np.diff(ptrs, append=len(global_locs)))
        return local_grp["local_locs"][()], col, global_locs


def _better_than_concat(arrs, cols, dtype):
    if not len(arrs):
        return np.empty((0, cols), dtype=dtype)
    cat = np.empty((sum(len(a) for a in arrs), cols), dtype=dtype)
    ptr = 0
    for arr in arrs:
        cat[ptr : ptr + len(arr)] = arr.reshape(-1, cols)
        ptr += len(arr)
    return cat


def get_dir_iter(direction):
    return ("inc", "out") if direction is None else (direction,)


class CSIterator:
    def __init__(self, cs, direction=None, local_=None, global_=None):
        self._cs = cs
        self._dir = direction
        self._lchunks = local_
        self._gchunks = global_

    def __iter__(self):
        if self._dir is None:
            yield from (
                (
                    direction,
                    CSIterator(self._cs, direction, self._lchunks, self._gchunks),
                )
                for direction in get_dir_iter(self._dir)
            )
        elif not isinstance(self._lchunks, Chunk):
            yield from (
                (lchunk, CSIterator(self._cs, self._dir, lchunk, self._gchunks))
                for lchunk in self.get_local_iter(self._dir, self._lchunks)
            )
        elif not isinstance(self._gchunks, Chunk):
            yield from (
                (
                    gchunk,
                    self._cs.load_block_connections(self._dir, self._lchunks, gchunk),
                )
                for gchunk in self.get_global_iter(
                    self._dir, self._lchunks, self._gchunks
                )
            )
        else:
            yield self._cs.load_block_connections(
                self._dir, self._lchunks, self._gchunks
            )

    def get_local_iter(self, direction, local_):
        if local_ is None:
            return self._cs.get_local_chunks(direction)
        elif isinstance(local_, Chunk):
            return (local_,)
        else:
            return iter(chunklist(local_))

    def get_global_iter(self, direction, local_, global_):
        if global_ is None:
            return self._cs.get_global_chunks(direction, local_)
        elif isinstance(global_, Chunk):
            return (global_,)
        else:
            return iter(chunklist(global_))


def _point_to_2d(arr):
    arr = np.array(arr, copy=False, dtype=int)
    if arr.ndim == 1:
        ret = np.full((len(arr), 3), -1)
        ret[:, 0] = arr
        return ret
    else:
        return arr



import json
import time
from uuid import uuid4

import numpy as np
from bsb import FileStore as IFileStore
from bsb import MissingActiveConfigError

from .resource import Resource

_root = "files"


class FileStore(Resource, IFileStore):
    def __init__(self, engine):
        super().__init__(engine, _root)

    def __bool__(self):
        return True

    def all(self):
        with self._engine._read():
            with self._engine._handle("r") as root:
                store = root[self._path]
                return {
                    id: json.loads(f.attrs.get("meta", "{}")) for id, f in store.items()
                }

    def load(self, id):
        with self._engine._read():
            with self._engine._handle("r") as root:
                ds = root[f"{self._path}/{id}"]
                data = ds[()]
                if encoding := ds.attrs.get("encoding", None):
                    data = data.decode(encoding)
                return data, json.loads(ds.attrs.get("meta", "{}"))

    def remove(self, id):
        with self._engine._write():
            with self._engine._handle("a") as root:
                del root[f"{self._path}/{id}"]

    def store(self, content, meta=None, id=None, encoding=None, overwrite=False):
        if id is None:
            id = str(uuid4())
        if meta is None:
            meta = {}
        meta["mtime"] = int(time.time())
        with self._engine._write():
            with self._engine._handle("a") as root:
                store = root[self._path]
                if isinstance(content, str):
                    if encoding is None:
                        encoding = "utf-8"
                    content = content.encode(encoding)
                content = np.array(content)
                if overwrite:
                    try:
                        del store[id]
                    except KeyError:
                        pass
                try:
                    ds = store.create_dataset(id, data=content)
                except ValueError:
                    raise Exception(f"File `{id}` already exists in store.")
                if encoding:
                    ds.attrs["encoding"] = encoding
                ds.attrs["meta"] = json.dumps(meta)
                ds.attrs["mtime"] = meta["mtime"]
        return id

    def load_active_config(self):
        """
        Load the active configuration stored inside of the storage.

        :returns: The active configuration that is loaded when this storage object is.
        :rtype: ~bsb.config.Configuration
        """
        from bsb.config import Configuration

        cfg_id = self._active_config_id()
        if cfg_id is None:
            raise MissingActiveConfigError("No active config")
        else:
            content, meta = self.load(cfg_id)
            # It's a serialized Python dict, so it should be JSON readable. We don't use
            # evaluate because files might originate from untrusted sources.
            tree = json.loads(content)
            cfg = Configuration(**tree)
            cfg._meta = meta
            return cfg

    def store_active_config(self, config):
        """
        Set the active configuration for this network.

        :param config: The active configuration that will be loaded when this storage
          object is.
        :type config: ~bsb.config.Configuration
        """
        id = self._active_config_id()
        self._engine.comm.barrier()
        if id is not None and self._engine.comm.get_rank() == 0:
            self.remove(id)
        config._meta["active_config"] = True
        active_id = None
        if self._engine.comm.get_rank() == 0:
            meta = {k: v for k, v in config._meta.items() if v is not None}
            active_id = self.store(json.dumps(config.__tree__()), meta)
        return self._engine.comm.bcast(active_id, root=0)

    def _active_config_id(self):
        match = (id for id, m in self.all().items() if m.get("active_config", False))
        return next(match, None)

    def has(self, id):
        with self._engine._read():
            with self._engine._handle("r") as root:
                return id in root

    def get_mtime(self, id):
        with self._engine._read():
            with self._engine._handle("r") as root:
                return root[self._path][id].attrs["mtime"]

    def get_encoding(self, id):
        with self._engine._read():
            with self._engine._handle("r") as root:
                return root[self._path][id].attrs.get("encoding", None)

    def get_meta(self, id):
        with self._engine._read():
            with self._engine._handle("r") as root:
                return json.loads(root[self._path][id].attrs.get("meta", "{}"))
